from .app import (main, __version__, g3d, g3d_by_inchi, check_functional_group, create_graph, compound, compound_by_cid,
                  compound_by_inchi, create_custom_functional_groups, count_functional_group, __author__, create_molecule_graph)

__all__ = ['main', '__version__', '__author__', 'g3d',
           'g3d_by_inchi', 'check_functional_group', 'create_graph', 'compound', 'compound_by_cid', 'compound_by_inchi', 'create_custom_functional_groups', 'count_functional_group', 'create_molecule_graph']



# import packages/modules
import os
from pathlib import Path
from networkx import Graph
import pubchemquery as pcq
import pandas as pd
from typing import List, Dict, Union, Literal

# internal
from .config import packageName
from .config import packageShortName
from .config import __version__
from .config import __description__
from .config import __author__
from .docs import MolParser, Compound, CustomChemGraph, Utility, Molecule


def main():
    # short description and version
    _des = f"{packageShortName} {__version__}: {__description__}"
    print(_des)


def compound(f: Union[str, Path]) -> Compound:
    '''
    Create a compound by parsing sdf file

    Parameters
    ----------
    f : str
        molecule file format (sdf) or string (sdf)

    Returns
    -------
    compound : object
        compound object
    '''
    try:
        # check str, path
        if not isinstance(f, (str, Path)):
            raise ValueError("Invalid input file path or string")

        # check file exists
        if os.path.exists(f):
            # parse file
            MolParserC = MolParser(f)
            compound_info = MolParserC.read_file()
            # compound
            compound = Compound(compound_info)
            # res
            return compound
        else:
            # check string
            if isinstance(f, str):
                # parse string
                MolParserC = MolParser(None)
                compound_info = MolParserC.read_file(
                    sourceContent={
                        'content': f.strip(),
                        'format': 'sdf'
                    })
                # compound
                compound = Compound(compound_info)
                # res
                return compound
            else:
                raise ValueError("Invalid input file path or string")
    except Exception as e:
        raise Exception(f"creating compound is failed! {e}")


def compound_by_cid(cid: Union[str, int]) -> Compound:
    '''
    Create a compound by cid

    Parameters
    ----------
    cid : str | int
        compound id

    Returns
    -------
    compound : object
        compound object
    '''
    try:
        # get sdf by cid
        sdf = pcq.get_structure_by_cid(str(cid))
        # check
        if sdf is None or len(sdf) == 0:
            # err
            raise Exception("sdf is not found!")
        # compound
        return compound(sdf)
    except Exception as e:
        raise Exception(f"creating compound is failed! {e}")


def compound_by_inchi(inchi: str) -> Compound:
    '''
    Create a compound by inchi

    Parameters
    ----------
    inchi : str
        compound inchi

    Returns
    -------
    compound : object
        compound object
    '''
    try:
        # get cid by inchi
        cid = pcq.get_cid_by_inchi(inchi)
        # check
        if cid:
            # get sdf by cid
            sdf = pcq.get_structure_by_cid(cid)
            # check
            if sdf is None or len(sdf) == 0:
                # err
                raise Exception("sdf is not found or has no content!")
            # compound
            return compound(sdf)
        else:
            # err
            raise Exception("inchi is not valid.")
    except Exception as e:
        raise Exception(f"creating compound is failed! {e}")


def create_graph(file: Path) -> Graph:
    '''
    Converts a sdf compound file to a graph

    Parameters
    ----------
    file : str
        molecule file format (sdf)

    Returns
    -------
    graph : object
        compound graph
    '''
    try:
        if os.path.exists(file):
            # parse file
            MolParserC = MolParser(file)
            compound_info = MolParserC.read_file()
            # compound
            compound = Compound(compound_info)
            # display 3d
            graph = compound.create_graph()
            # res
            return graph
        else:
            raise Exception("file path is not valid.")
    except Exception as e:
        raise Exception(f"creating graph is failed! {e}")


def g3d(f: Union[str, Path], fig_size: List = [], bg_color: str = '#ffffff',
        display_legend: bool = True, display_atom_id: bool = True,
        display_bond_length: bool = False):
    '''
    3d graph of a compound

    Parameters
    ----------
    f : str
        molecule file format (sdf) or a sdf string variable
    fig_size : list
        figure size (default [])
    bg_color : str
        background color (default '#ffffff')
    display_legend : bool
        display legend (default True)
    display_atom_id : bool
        display atom id (default True)
    display_bond_length : bool
        display bond length (default True)

    Returns
    -------
    None
        display 3d graph
    '''
    try:
        # create a compound
        comp = compound(f)
        # display 3d
        comp.view3d(fig_size=fig_size, display_legend=display_legend, bg_color=bg_color,
                    display_atom_id=display_atom_id, display_bond_length=display_bond_length)
    except Exception as e:
        raise Exception(f"file path/variable is not valid! {e}")


def g3d_by_inchi(inchi: str, fig_size: List = [], bg_color: str = '#ffffff', display_legend: bool = True, display_atom_id: bool = True, display_bond_length: bool = False):
    '''
    3d graph of a compound using its InChI identifier

    Parameters
    ----------
    inchi : str
        inchi code
    fig_size : list
        figure size (default [])
    bg_color : str
        background color (default '#ffffff')
    display_legend : bool
        display legend (default True)
    display_atom_id : bool
        display atom id (default True)
    display_bond_length : bool
        display bond length (default True)

    Returns
    -------
    None
        display 3d graph
    '''
    # check inchi
    if inchi is not None:
        # get cid
        cid = pcq.get_cid_by_inchi(inchi)
        # check
        if cid is not None:
            # get sdf
            sdf = pcq.get_structure_by_cid(cid)
            # check
            if sdf is not None:
                # display 3d
                # parse file
                MolParserC = MolParser(None)
                compound_info = MolParserC.read_file(
                    sourceContent={
                        'content': sdf,
                        'format': 'sdf'
                    })
                # compound
                compound = Compound(compound_info)
                # display 3d
                compound.view3d(fig_size=fig_size, display_legend=display_legend, bg_color=bg_color,
                                display_atom_id=display_atom_id, display_bond_length=display_bond_length)
            else:
                raise Exception("sdf is not found!")
        else:
            raise Exception("cid is not found!")
    else:
        raise Exception("inchi is not valid.")


def check_functional_group(file: Union[str, Path], functional_groups: List[Union[str, CustomChemGraph]] = [], res_format: Literal['original', 'dataframe'] = 'original'):
    '''
    Check a functional group exists in a compound

    Parameters
    ----------
    file : Path | str
        molecule file format (sdf) or string (sdf)
    functional_groups : list[str] or CustomChemGraph object
        functional group (default ['hydroxyl']) or CustomChemGraph object
    res_format : str
        result format (default 'original')

    Returns
    -------
    res : dict
        a list of all count
    compound : Compound
        compound object (sdf file)
    '''
    try:
        # create compound
        comp = compound(file)

        # check functional group
        res = comp.check_functional_groups(functional_groups)

        # check
        if res_format == 'dataframe':
            # dataframe
            df = pd.DataFrame(res)
            return df, comp
        elif res_format == 'original':
            # raw
            return res, comp
        else:
            raise Exception("res_format is not valid.")

    except Exception as e:
        raise Exception(f"checking functional group is failed! {e}")


def count_functional_group(file: Union[str, Path], functional_groups: List[Union[str, CustomChemGraph]] = [], res_format: Literal['original', 'dataframe'] = 'original'):
    '''
    Counts the occurrences of functional groups within the structure of a compound.

    Parameters
    ----------
    file : Path | str
        molecule file format (sdf) or string (sdf)
    functional_groups : list[str] or CustomChemGraph object
        functional group (default ['hydroxyl']) or CustomChemGraph object
    res_format : str
        result format (default 'original')

    Returns
    -------
    res : dict
        a list of all count
    compound : Compound
        compound object (sdf file)
    '''
    try:
        # create compound
        comp = compound(file)

        # create graph
        comp.create_graph()
        # check functional group
        res = comp.check_functional_groups(
            functional_groups, count_functional_group=True)
        # check
        if res_format == 'dataframe':
            # dataframe
            df = pd.DataFrame(res)
            return df, comp
        elif res_format == 'original':
            # raw
            return res, comp
        else:
            raise Exception("res_format is not valid.")

    except Exception as e:
        raise Exception(f"counting functional group is failed! {e}")


def create_custom_functional_groups(functional_groups: Union[Dict[str, List[str]], List[Dict[str, List[str]]], Path, str]) -> CustomChemGraph:
    '''
    Creates custom functional groups based on the following example.

    Parameters
    ----------
    functional_groups : list[dict] or dict
        functional group

    Returns
    -------
    custom_functional_groups : list[dict]
        a list of all custom functional group

    Examples
    --------
    ```python
    # fg1: CH2-O
    # fg2: CH2CHO
    # List of custom functional groups
    custom_functional_group = [
        {'fg1': ["C1-H1","C1-H2","C1-O1"]},
        {'fg2': ["C1-H1","C1-H2","C1-C2","C2-H3","C2-O2"]}
    ]

    # Dict of custom functional groups
    custom_functional_group = {
        'fg1': ["C1-H1","C1-H2","C1-O1"],
        'fg2': ["C1-H1","C1-H2","C1-C2","C2-H3","C2-O2"]
    }
    ```
    '''
    try:
        # check format
        if isinstance(functional_groups, dict):
            # set a list
            if len(functional_groups) == 1:
                custom_functional_groups = [functional_groups]
            else:
                custom_functional_groups = [
                    {k: v} for k, v in functional_groups.items()
                ]
        elif isinstance(functional_groups, list):
            # check is a list of dict
            if all(isinstance(item, dict) for item in functional_groups):
                # set
                custom_functional_groups = functional_groups
        elif isinstance(functional_groups, (str, Path)):
            # check is a file yml
            custom_functional_groups = Utility.load_custom_functional_group(
                functional_groups)
        else:
            raise Exception("functional_groups format is not valid.")

        # custom chem graph
        CustomChemGraphC = CustomChemGraph(custom_functional_groups)
        # res
        return CustomChemGraphC
    except Exception as e:
        raise Exception(f'creating custom functional group is failed! {e}')


def create_molecule_graph(molecule_src: Union[Dict[str, List[str]], Path], molecule_name: str = 'MainChain') -> CustomChemGraph:
    '''
    Make a molecule graph from custom functional groups based on the following example.

    Parameters
    ----------
    molecule_src : Dict[str, List[str]] | str
        molecule source
    molecule_name : str
        molecule name (default 'MainChain')

    Returns
    -------
    molecule : Molecule
        molecule object

    Examples
    --------
    ```python
    # molecule source
    molecule_src = {
        'MainChain': ["C1-C2","C2-C3","C3*{Chain1}","C3-C4","C4*{Chain2}","C4-C5","C5-C6"],
        'Chain1': ["C1=C2","C2-C3","C3=*"],
        'Chain2' : ["*-C1","C1=C2","C2-XX3"]
    }
    ```

    Notes
    -----
    - `*` is a connection point
    - `{}` is a reference to another chain
    - C3*{Chain1} means C3 is connected to Chain1 from C3=*
    - Bond index starts from 1 in all chains
    - Bond type: single bond (-), double bond (=), triple bond (#)
    '''
    try:
        # check format
        if isinstance(molecule_src, (str, Path)):
            # check is a file yml
            _mol_src = Utility.load_custom_functional_group(
                molecule_src)

        # init molecule
        MoleculeC = Molecule(molecule_src, molecule_name)

        # build molecule
        _, _, _, molecule = MoleculeC.build()

        # custom chem graph
        CustomChemGraphC = CustomChemGraph([molecule])

        # res
        return CustomChemGraphC
    except Exception as e:
        raise Exception(f'creating custom functional group is failed! {e}')



# NETWORK
# -------

# import packages/modules
import pandas as pd
import networkx as nx
from networkx.algorithms import isomorphism
# local
from .chemgraphs import ChemGraphs
from .customchemgraph import CustomChemGraph


class Network(ChemGraphs):

    # functional groups
    _functional_groups = []
    _custom_functional_groups = []
    _custom_functional_group_list = {}
    # graph
    _compound_graph = None

    def __init__(self, atomElements, atomBonds, xyzList, xyzCenterList, atomBonds1d):
        self.atomElements = atomElements
        # bond block (info)
        self.atomBonds = atomBonds
        self.xyzList = xyzList
        self.xyzCenterList = xyzCenterList

        # 1d vector of atom bonds
        self.atomBonds1d = atomBonds1d

        # TODO: super
        ChemGraphs.__init__(self)

        # functional group list
        self.function_group_list = {
            'hydroxyl': [self.graph_hydroxyl()],
            'carbonyl': [self.graph_carbonyl()],
            'carboxyl': [self.graph_carboxyl()],
            'N-H': [self.graph_N_H()],
            'C-N': [self.graph_C_N_single_bond()],
            'N-O': [self.graph_N_O_single_bond()],
            'C-O': [self.graph_C_O_single_bond()],
            'C#N': [self.graph_C_N_triple_bond()],
            'methyl-group': [self.graph_methyl()],
            'methylene-group': [self.graph_methylene()],
            'methine-group': [self.graph_methine()],
            'ether': [self.graph_ether()],
            'pst-alcohols': [self.graph_primary_alcohol(), self.graph_secondary_alcohol(),
                             self.graph_tertiary_alcohol(), self.graph_secondary_alcohol_double_bond(),
                             self.graph_primary_alcohol_double_bond()],
            'primary-alcohol': [self.graph_primary_alcohol(), self.graph_primary_alcohol_double_bond()],
            'secondary-alcohol': [self.graph_secondary_alcohol(), self.graph_secondary_alcohol_double_bond()],
            'tertiary-alcohol': [self.graph_tertiary_alcohol()],
            'alkane': [self.graph_alkane()],
            'alkane CH bond (sp3)': [self.graph_alkane_CH_bond()],
            'alkene': [self.graph_alkene()],
            'alkene CH bond (sp2)': [self.graph_alkene_CH_bond()],
            'aromatic CH bond (sp2)': [self.graph_aromatic_CH_bond()],
            'aldehyde CH bond (sp2)': [self.graph_aldehyde_CH_bond()],
            'alkyne': [self.graph_alkyne()],
            'alkyne CH bond (sp)': [self.graph_alkyne_CH_bond()],
            'arene': [self.graph_arene()],
            'aldehyde': [self.graph_aldehyde()],
            'ketone': [self.graph_ketone()],
            'carboxylic-acid': [self.graph_carboxylic_acid()],
            'ester': [self.graph_ester()],
            'pst-amide': [self.graph_primary_amide(), self.graph_secondary_amide(), self.graph_tertiary_amide()],
            'primary-amide': [self.graph_primary_amide()],
            'secondary-amide': [self.graph_secondary_amide()],
            'tertiary-amide': [self.graph_tertiary_amide()],
            'pst-amine': [self.graph_primary_amine(), self.graph_secondary_amine(), self.graph_tertiary_amine()],
            'primary-amine': [self.graph_primary_amine()],
            'secondary-amine': [self.graph_secondary_amine()],
            'tertiary-amine': [self.graph_tertiary_amine()],
            'nitrile': [self.graph_nitrile()],
            'thiol': [self.graph_thiol()],
            'alkyl-halids': [self.graph_alkyl_halide('F'), self.graph_alkyl_halide('Cl'), self.graph_alkyl_halide('Br'),
                             self.graph_alkyl_halide('I'),
                             self.graph_primary_alkyl_halide(
                                 'F'), self.graph_primary_alkyl_halide('Cl'),
                             self.graph_primary_alkyl_halide('Br'), self.graph_primary_alkyl_halide('I')],
            'epoxide': [self.graph_epoxide()]
        }

        # update functional groups
        self.functional_groups = [i for i in self.function_group_list.keys()]

    # property
    @property
    def functional_groups(self):
        return self._functional_groups

    # setter
    @functional_groups.setter
    def functional_groups(self, value):
        self._functional_groups = []
        self._functional_groups = [*value]

    @property
    def custom_functional_groups(self):
        return self._custom_functional_groups

    @custom_functional_groups.setter
    def custom_functional_groups(self, value):
        self._custom_functional_groups = []
        self._custom_functional_groups = [*value]

    @property
    def custom_functional_group_list(self):
        return self._custom_functional_group_list

    @custom_functional_group_list.setter
    def custom_functional_group_list(self, value):
        self._custom_functional_group_list = value

    @property
    def compound_graph(self):
        return self._compound_graph

    @compound_graph.setter
    def compound_graph(self, value):
        self._compound_graph = value

    def check_functional_groups(self, functional_groups=[],
                                count_functional_group=False):
        '''
        Check functional groups in a compound

        Parameters
        ----------
        functional_groups : list
            list of functional groups

        Returns
        -------
        res : dict
            a list of all count
        '''
        # check functional group
        if len(functional_groups) == 0:
            functional_groups = list(self.function_group_list.keys())

        # create graph
        G = self.create_graph()

        # check
        if count_functional_group:
            res = self.count_functional_group(G, functional_groups)
        else:
            # check functional groups
            res = self.check_functional_group(G, functional_groups)
        # res
        return res

    def check_functional_group(self, G, function_groups):
        '''
        Check a functional group exists in a compound

        Parameters
        ----------
        G : graph
            graph
        function_groups : list[str]
            functional group name like hydroxyl

        Returns
        -------
        res : dict
            a list of all count
        '''
        def node_match(n1, n2):
            '''
            Define a custom node_match function to ensure element matching
            '''
            # Define dummy node pattern
            dummy_node_pattern = "XX"

            # Ignore dummy nodes
            if n1['symbol'].startswith(dummy_node_pattern) or n2['symbol'].startswith(dummy_node_pattern):
                return True

            return n1['symbol'] == n2['symbol']

        def edge_match(e1, e2):
            '''
            Define a custom edge_match function to ensure bond type matching
            '''
            return e1['type'] == e2['type']

        # res
        res_match = []

        # NOTE: for each functional group
        for item in function_groups:
            # * Check instance
            if isinstance(item, str):
                # ! check functional exists in the list
                if item in self.function_group_list:
                    # get a list of graphs
                    # REVIEW
                    function_group_graphs = self.function_group_list[item]

                    # flag to track if functional group is found
                    fg_found_any = False

                    # graph function list
                    for _fn in function_group_graphs:
                        # Create a GraphMatcher object for a functional group
                        fg_matcher = isomorphism.GraphMatcher(
                            G, _fn, node_match=node_match,
                            edge_match=edge_match)

                        # Check if a functional group is in the main graph
                        fg_found = fg_matcher.subgraph_is_isomorphic()

                        if fg_found:
                            # print(f"{item} found in the molecule!")
                            fg_found_any = True
                            # res
                            res_match.append({
                                'function_group': item,
                                'result': fg_found
                            })
                            # break
                            break
                    # check
                    if not fg_found_any:
                        # print(f"{item} not found in the molecule.")
                        # res
                        res_match.append({
                            'function_group': item,
                            'result': False
                        })
            elif isinstance(item, CustomChemGraph):
                # get a list of graphs
                for custom_functional_group in item.custom_functional_groups:
                    # check dict
                    if len(custom_functional_group) == 1:
                        # get key (custom functional group name)
                        key = list(custom_functional_group.keys())[0]
                        # get value (graph)
                        value = custom_functional_group[key]
                        # print(type(value))

                        # ANCHOR: check value type
                        if isinstance(value, nx.Graph):

                            # update custom functional group
                            self.update_custom_functional_group(key, value)

                            # Create a GraphMatcher object for a functional group
                            fg_matcher = isomorphism.GraphMatcher(
                                G, value, node_match=node_match, edge_match=edge_match)

                            # Check if a functional group is in the main graph (bool)
                            fg_found = fg_matcher.subgraph_is_isomorphic()
                            # print(f"{key} found in the molecule!")

                            # res
                            res_match.append({
                                'function_group': key,
                                'result': fg_found
                            })

                        elif isinstance(value, list):
                            # SECTION: list of graphs {already created}

                            # graphs set
                            graphs_set_ = self.__look_for_group_subgroup(value)

                            # NOTE: start matching
                            if graphs_set_['group'] is not None and graphs_set_['subgroup'] is not None:
                                # Create a GraphMatcher object for a functional group
                                fg_matcher = isomorphism.GraphMatcher(
                                    G, graphs_set_['group'], node_match=node_match, edge_match=edge_match)

                                # Check if a functional group is in the main graph (bool)
                                fg_found = fg_matcher.subgraph_is_isomorphic()

                                # SECTION: group found
                                if fg_found:
                                    # group sections
                                    group_sections = list(
                                        fg_matcher.subgraph_isomorphisms_iter())

                                    # remove duplicated group sections
                                    group_sections_filtered = self.remove_duplicated_subgraphs(
                                        group_sections)

                                    if len(group_sections_filtered) == 0:
                                        raise Exception('group not found!')

                                    # counter
                                    fg_subgroups_num = 0
                                    # looping through group sections
                                    for gs_filtered in group_sections_filtered:

                                        # create group pattern within the main graph
                                        group_pattern = G.subgraph(
                                            gs_filtered.keys())

                                        # Create a GraphMatcher object for a functional group
                                        fg_matcher_subgroup = isomorphism.GraphMatcher(
                                            group_pattern, graphs_set_['subgroup'], node_match=node_match, edge_match=edge_match)

                                        # Check if a functional group is in the main graph (bool)
                                        fg_found_subgroup = fg_matcher_subgroup.subgraph_is_isomorphic()

                                        # check
                                        if fg_found_subgroup:
                                            fg_subgroups_num += 1

                                    # check
                                    if fg_subgroups_num > 0:
                                        # res
                                        res_match.append({
                                            'function_group': key,
                                            'result': fg_found_subgroup
                                        })
                                    else:
                                        # res
                                        res_match.append({
                                            'function_group': key,
                                            'result': False
                                        })
                                else:
                                    # res
                                    res_match.append({
                                        'function_group': key,
                                        'result': False
                                    })

        # REVIEW
        # get a list of functional group names
        function_group_names = [x['function_group'] for x in res_match]
        # update list
        self.functional_groups = function_group_names

        # res
        return res_match

    def count_functional_group(self, G, function_groups):
        '''
        Count the occurrences of functional groups within the structure of a compound.

        Parameters
        ----------
        G : graph
            graph
        function_groups : list[str]
            functional group name like hydroxyl

        Returns
        -------
        res : dict
            a list of all count
        '''
        def node_match(n1, n2):
            '''
            Define a custom node_match function to ensure element matching
            '''
            # Define dummy node pattern
            dummy_node_pattern = "XX"

            # Ignore dummy nodes
            if n1['symbol'].startswith(dummy_node_pattern) or n2['symbol'].startswith(dummy_node_pattern):
                return True

            return n1['symbol'] == n2['symbol']

        def edge_match(e1, e2):
            '''
            Define a custom edge_match function to ensure bond type matching
            '''
            return e1['type'] == e2['type']

        # res
        res_match = []

        # for each functional group
        for item in function_groups:
            # * Check instance
            if isinstance(item, str):
                # ! check functional exists in the list
                if item in self.function_group_list:
                    # get a list of graphs
                    # REVIEW
                    function_group_graphs = self.function_group_list[item]

                    # flag to track if functional group is found
                    fg_found_any = False
                    # count of functional group
                    fg_count = 0
                    seen_subgraphs = set()

                    # graph function list
                    for _fn in function_group_graphs:
                        # Create a GraphMatcher object for a functional group
                        fg_matcher = isomorphism.GraphMatcher(
                            G, _fn, node_match=node_match, edge_match=edge_match)

                        # Check if a functional group is in the main graph
                        # fg_found = fg_matcher.subgraph_is_isomorphic()

                        # ! Check if a functional group is in the main graph
                        for subgraph in fg_matcher.subgraph_isomorphisms_iter():
                            # Convert the subgraph to a canonical form
                            canonical_subgraph = tuple(
                                sorted(subgraph.keys()))

                            # Check if the subgraph has been seen before
                            if canonical_subgraph not in seen_subgraphs:
                                seen_subgraphs.add(canonical_subgraph)
                                fg_found_any = True
                                fg_count += 1

                    # check
                    if fg_found_any:
                        # print(f"{item} found in the molecule!")
                        res_match.append({
                            'function_group': item,
                            'result': True,
                            'count': fg_count
                        })
                    else:
                        # print(f"{item} not found in the molecule.")
                        res_match.append({
                            'function_group': item,
                            'result': False,
                            'count': 0
                        })

            elif isinstance(item, CustomChemGraph):  # ! custom functional groups
                # get a list of graphs
                for custom_functional_group in item.custom_functional_groups:
                    # check dict
                    if len(custom_functional_group) == 1:

                        # flag to track if functional group is found
                        fg_found_any = False
                        # count of functional group
                        fg_count = 0
                        seen_subgraphs = set()

                        # get key (custom functional group name)
                        key = list(custom_functional_group.keys())[0]
                        # get value (graph)
                        value = custom_functional_group[key]

                        # ANCHOR: check value type
                        # SECTION: single graph
                        if isinstance(value, nx.Graph):

                            # update custom functional group
                            self.update_custom_functional_group(key, value)

                            # Create a GraphMatcher object for a functional group
                            fg_matcher = isomorphism.GraphMatcher(
                                G, value, node_match=node_match, edge_match=edge_match)

                            # Check if a functional group is in the main graph
                            # fg_found = fg_matcher.subgraph_is_isomorphic()
                            # print(f"{key} found in the molecule!")

                            # ! Check if a functional group is in the main graph
                            for subgraph in fg_matcher.subgraph_isomorphisms_iter():
                                # Convert the subgraph to a canonical form
                                canonical_subgraph = tuple(
                                    sorted(subgraph.keys()))

                                # Check if the subgraph has been seen before
                                if canonical_subgraph not in seen_subgraphs:
                                    seen_subgraphs.add(canonical_subgraph)
                                    fg_found_any = True
                                    fg_count += 1

                            # check
                            if fg_found_any:
                                # print(f"{item} found in the molecule!")
                                res_match.append({
                                    'function_group': key,
                                    'result': True,
                                    'count': fg_count
                                })
                            else:
                                # print(f"{item} not found in the molecule.")
                                res_match.append({
                                    'function_group': key,
                                    'result': False,
                                    'count': 0
                                })
                        # SECTION: list of graphs {already created}
                        elif isinstance(value, list):
                            # graphs set
                            graphs_set_ = self.__look_for_group_subgroup(value)

                            # NOTE: start matching
                            if graphs_set_['group'] is not None and graphs_set_['subgroup'] is not None:
                                # Create a GraphMatcher object for a functional group
                                fg_matcher = isomorphism.GraphMatcher(
                                    G, graphs_set_['group'], node_match=node_match, edge_match=edge_match)

                                # Check if a functional group is in the main graph (bool)
                                fg_found = fg_matcher.subgraph_is_isomorphic()

                                # SECTION: group found
                                if fg_found:
                                    # group sections
                                    group_sections = list(
                                        fg_matcher.subgraph_isomorphisms_iter())

                                    # remove duplicated group sections
                                    group_sections_filtered = self.remove_duplicated_subgraphs(
                                        group_sections)

                                    if len(group_sections_filtered) == 0:
                                        raise Exception('group not found!')

                                    # counter
                                    fg_subgroups_num = 0
                                    # looping through group sections
                                    for gs_filtered in group_sections_filtered:

                                        # create group pattern within the main graph
                                        group_pattern = G.subgraph(
                                            gs_filtered.keys())

                                        # Create a GraphMatcher object for a functional group
                                        fg_matcher_subgroup = isomorphism.GraphMatcher(
                                            group_pattern, graphs_set_['subgroup'], node_match=node_match, edge_match=edge_match)

                                        # Check if a functional group is in the main graph (bool)
                                        fg_found_subgroup = fg_matcher_subgroup.subgraph_is_isomorphic()

                                        # check
                                        if fg_found_subgroup:
                                            fg_subgroups_num += 1

                                    # check
                                    if fg_subgroups_num > 0:
                                        # res
                                        res_match.append({
                                            'function_group': key,
                                            'result': fg_found_subgroup,
                                            'count': fg_subgroups_num
                                        })
                                    else:
                                        # res
                                        res_match.append({
                                            'function_group': key,
                                            'result': False,
                                            'count': 0

                                        })
                                else:
                                    # res
                                    res_match.append({
                                        'function_group': key,
                                        'result': False,
                                        'count': 0
                                    })

        # get a list of functional group names
        function_group_names = [x['function_group'] for x in res_match]
        # update list
        self.functional_groups = function_group_names

        # res
        return res_match

    def create_graph(self):
        '''
        Check functional groups in a compound

        Parameters
        ----------
        functional_groups : list
            list of functional groups

        Returns
        -------
        res : dict
            a list of all count
        '''
        # atom no
        atomNo = len(self.xyzList)
        # bond no
        bondNo = len(self.atomBonds1d)

        # Create a graph from atoms and bonds
        G = nx.Graph()

        # *** atom visualization
        for i in range(atomNo):
            # xyz
            _atom1X = self.xyzList[i, 0]
            _atom1Y = self.xyzList[i, 1]
            _atom1Z = self.xyzList[i, 2]
            _atom1XYZ = [_atom1X, _atom1Y, _atom1Z]

            # color
            # atom id
            _atomId = int(i+1)
            # symbol
            _atomSymbol = str(self.atomElements[i]).strip()
            # atom mark
            atomMark = str(_atomSymbol) + str(_atomId)

            # add node with cooridination
            # G.add_node(_atomId, symbol=_atomSymbol,
            #            x=_atom1X, y=_atom1Y, z=_atom1Z, xyz=_atom1XYZ)

            G.add_node(_atomId, symbol=_atomSymbol)

        # *** using bond block

        for i in range(bondNo):
            # atom 1
            _id1 = self.atomBonds1d[i]['id1']
            _symbol1 = self.atomBonds1d[i]['symbol1']
            # atom 2
            _id2 = self.atomBonds1d[i]['id2']
            _symbol2 = self.atomBonds1d[i]['symbol2']
            # bond type
            _bondType = self.atomBonds1d[i]['bond_type']
            # bond symbol
            _bondSymbol = self.atomBonds1d[i]['bond_symbol']

            # add edge
            G.add_edge(_id1, _id2, symbol=_bondSymbol, type=_bondType)

        # update
        self.compound_graph = G

        # res
        return G

    def node_match(self, n1, n2):
        '''
        Define a custom node_match function to ensure element matching
        '''
        # Define dummy node pattern
        dummy_node_pattern = "XX"

        # Ignore dummy nodes
        if n1['symbol'].startswith(dummy_node_pattern) or n2['symbol'].startswith(dummy_node_pattern):
            return True

        return n1['symbol'] == n2['symbol']

    def edge_match(self, e1, e2):
        '''
        Define a custom edge_match function to ensure bond type matching
        '''
        return e1['type'] == e2['type']

    def search_within_main_graph(self, functional_group):
        '''
        Search a small graph within a larger graph

        Parameters
        ----------
        functional_group : str
            functional group
        '''

        # get the graph
        G = self.compound_graph
        # sub graph
        sub_graphs = []
        sub_graphs_name = []
        # res
        res_match = []
        fg_count = 0

        # search graph in functional_groups
        for item in self.functional_groups:
            if functional_group == item:
                sub_graphs_name = [
                    i for i in self.function_group_list.keys() if i == item]
                if len(sub_graphs_name) == 1:
                    sub_graphs = self.function_group_list[str(item)]
        # search graph in custom functional_groups
        for item in self.custom_functional_groups:
            if functional_group == item:
                sub_graphs_name = [
                    i for i in self.custom_functional_group_list.keys() if i == item]
                if len(sub_graphs_name) == 1:
                    sub_graphs = self.custom_functional_group_list[str(item)]

        # check
        if len(sub_graphs) == 0:
            raise Exception('functional group not found!')

        # Create a GraphMatcher object for a functional group
        for _fn in sub_graphs:
            fg_matcher = isomorphism.GraphMatcher(
                G, _fn, node_match=self.node_match, edge_match=self.edge_match)

            # Check if a functional group is in the main graph
            for subgraph in fg_matcher.subgraph_isomorphisms_iter():
                fg_count += 1

                # Store the subgraph
                # subgraph_nodes = subgraph.items()
                # pattern subgraph
                subgraph_pattern = G.subgraph(subgraph.keys())

                # Add the subgraph and node IDs to the result
                res_match.append({
                    'function_group': _fn,
                    'result': True,
                    'count': fg_count,
                    'subgraph': subgraph,
                    'subgraph_pattern': subgraph_pattern
                })

        return res_match

    def update_custom_functional_group(self, key, value):
        '''
        Update custom functional group
        '''
        # update custom functional group
        # check key exist
        if key not in self.custom_functional_groups:
            # save
            self.custom_functional_groups.append(key)
            # dict (key: graph)
            self.custom_functional_group_list[key] = [value]

    # NOTE: search deeper within the main graph
    def __look_for_group_subgroup(self, value) -> dict[str, nx.Graph]:
        """
        Look for a group and subgroup within the custom functional groups.

        Parameters
        ----------
        value : list
            list of graphs in a custom functional group (group and subgroup)

        Returns
        -------
        graphs_set_ : dict
            a dictionary of group and subgroup graphs
        """
        # check list (2 elements)
        if len(value) != 2:
            raise Exception(
                'custom functional group must have 2 elements')

        # elements
        graphs_ = {}

        # looping through value
        for v in value:
            # graph nodes number
            nodes_num = len(v.nodes)
            # graph edges number
            edges_num = len(v.edges)
            # graph name
            name = v.name
            # save
            graphs_[name] = {
                'nodes_num': nodes_num,
                'edges_num': edges_num,
                'graph': v
            }

        # set group and subgroup
        graphs_set_ = {}

        # reference
        ref_ = max([v['nodes_num']
                    for k, v in graphs_.items()])

        # looping through graphs_
        for k, v in graphs_.items():
            # check group
            if v['nodes_num'] >= ref_:
                graphs_set_['group'] = v['graph']
            else:
                graphs_set_['subgroup'] = v['graph']

        # check
        if graphs_set_['group'] is None:
            raise Exception('group not found!')

        # check
        if graphs_set_['subgroup'] is None:
            raise Exception('subgroup not found!')

        # return
        return graphs_set_

    def remove_duplicated_subgraphs(self, subgraphs):
        """
        Remove duplicated subgraphs.

        Parameters
        ----------
        subgraphs : list
            list of subgraphs
        """
        try:
            # Create a set to store tuples of dictionaries
            seen = set()

            # Filter out dictionaries with equal keys in different orders
            filtered_data = []
            for i, d in enumerate(subgraphs):
                # print(f"d{i+1}: {d}")
                # print(f"d keys: {d.keys()}")
                # print(f"d sorted keys: {sorted(d.keys())}")
                # print(f"d values: {d.values()}")
                # print("-"*5)

                # Convert dictionary to tuple of items (key, value pairs), sorted by keys
                sorted_tuple = tuple(sorted(d.keys()))

                # Check if the tuple has been seen before
                if sorted_tuple not in seen:
                    filtered_data.append(d)
                    seen.add(sorted_tuple)

            # return
            return filtered_data

        except Exception as e:
            raise Exception(f"Remove duplicated subgraphs error: {e}")



# UTILITY CLASS
# ---------------

# import libs
import os
import json
import fnmatch
import re
import csv
import random
import numpy as np
from datetime import date
from random import randint
import yaml
from typing import List, Dict


class Utility():

    json_csv_columns = [
        'file_name',
        'image_name_1',
        'image_name_2',
        'image_name_3',
        'mat_structure',
        'atom_numbers',
        'mat_cid',
        'mat_name',
        'mat_formula',
        'mat_smiles',
        'mat_mass',
        'atom_elements',
        'atom_atomic_number',
        'bond_numbers',
        'bond_list',
        'xyz_list',
        'xyz_center_list',
        'atom_weights',
        'mat_inchikey'
    ]

    def __init__(self):
        pass

    @staticmethod
    def CheckFileFormat(filePath):
        '''
        check file format

        args:
            filePath: file name dir

        return:
            file directory, file name, file format
        '''
        # check file exist
        if os.path.isfile(filePath):
            # file analysis
            fileDir = os.path.dirname(filePath)
            fileName = os.path.basename(filePath)
            fileFormat = os.path.splitext(filePath)[1]
            fileFormat = str(fileFormat.split(".")[-1]).lower()
            # res
            return fileDir, fileName, fileFormat
        else:
            raise Exception('file path is not valid.')

    @staticmethod
    def ReadContent(sourceContent):
        '''
        Read (load) file with respect to its format extension

        Parameters
        ----------
        sourceContent : dict
            contentFile: string content of a text file (url request)
            contentFormat: content format (sdf, json, ...)

        Returns
        -------
        file content : str
            content of a text file
        _contentFormat : str
            content format
        '''
        try:
            #
            _contentSource, _contentFormat = sourceContent.values()
            # check
            if _contentSource and _contentFormat:
                # read file
                if _contentFormat == 'sdf':
                    # string
                    fileContent = _contentSource
                elif _contentFormat == 'json':
                    # json convert to dict
                    fileContent = json.loads(_contentSource)
                elif _contentFormat == 'json-string':
                    # json convert to dict
                    fileContent = json.loads(_contentSource)

                # res
                return fileContent, _contentFormat
            else:
                raise Exception("target path is not valid.")
        except Exception as e:
            raise Exception(e)

    @staticmethod
    def OpenFile(filePath):
        '''
        Open file with respect to its format extension

        Parameters
        ----------
        filePath : str
            file path

        Returns
        -------
        file content : str
            content of a text file
        fileDir : str
            file directory
        fileName : str
            file name
        fileFormat : str
            file format
        '''
        try:
            # check
            if os.path.exists(filePath):
                # file info
                fileDir, fileName, fileFormat = Utility.CheckFileFormat(
                    filePath)

                # read a file
                with open(filePath, 'r') as f:
                    if fileFormat == 'sdf':
                        fileContent = f.read()
                    elif fileFormat == 'json':
                        fileContent = json.load(f)

                # res
                return fileContent, fileDir, fileName, fileFormat
            else:
                raise Exception("target path is not valid.")

        except Exception as e:
            raise Exception(e)

    @staticmethod
    def ListFiles(targetPath, fileExtension=''):
        '''
        list files in a target file

        Parameters
        ----------
        targetPath : str
            target path
        fileExtension : str
            file extension, default is empty

        Returns
        -------
        filesFound : list
            list of files in the target path
        '''
        try:
            # check
            if os.path.exists(targetPath):
                if not fileExtension:
                    # files
                    filesFound = os.listdir(targetPath)
                else:
                    # res
                    filesFound = []
                    for f in os.listdir(targetPath):
                        if f.endswith('.'+str(fileExtension).strip()):
                            filesFound.append(f)
                # res
                return filesFound
            else:
                raise Exception("target path is not valid.")

        except Exception as e:
            raise Exception(e)

    @staticmethod
    def SaveFile(fileContent, fileName, fileFormat, fileDir, logMessage=' file is successfully created and saved in'):
        '''
        save a file with respect to a format

        Parameters
        ----------
        fileContent : str
            file content
        fileName : str
            file name
        fileFormat : str
            file format
        fileDir : str
            file directory
        logMessage : str
            log message

        Returns
        -------
        bool
            True if the file is successfully created and saved in
        '''
        try:
            # set
            _fileFormat = str(fileFormat).lower()
            # file full name with format
            _fileFullName = fileName + f'.{_fileFormat}'
            # file full location
            fileLoc = os.path.join(fileDir, _fileFullName)

            # check
            if not os.path.isdir(fileDir):
                raise Exception('file directory is not valid.')

            # open file
            with open(fileLoc, 'w') as f:
                # check
                if _fileFormat == 'sdf':
                    # save
                    f.write(fileContent)
                    f.close()
                # FIXME
                if _fileFormat == 'json-string':
                    # save
                    json.dumps(fileContent, f, indent=5)
                    f.close()
                if _fileFormat == 'json':
                    # save ()
                    json.dump(fileContent, f, indent=5)
                    f.close()

            # log
            print(f"the {_fileFormat + logMessage} `{fileLoc}`")

            return True
        except Exception as e:
            raise Exception(e)

    @staticmethod
    def FindTargetFileInFolder(file_location, file_format):
        '''
        Find target files with respect to a specific format such as json

        Parameters
        ----------
        file_location : str
            file location
        file_format : str
            file format

        Returns
        -------
        listFiles : list
            list of target files
        '''
        try:
            # access directory
            allFiles = os.listdir(file_location)
            # file format
            fileExtension = f"*.{file_format}"
            # list
            listFiles = fnmatch.filter(allFiles, fileExtension)
            # res
            return listFiles
        except Exception as e:
            raise Exception(e)

    def SelectFile(list_file_names, file_name_ids, file_name_prefix='cid'):
        '''
        select files from a file list based on a query

        Parameters
        ----------
        list_file_names : list
            list of file names
        file_name_ids : list
            list of file ids
        file_name_prefix : str
            file name prefix

        Returns
        -------
        listFiles : list
            list of selected files
        '''
        try:
            # res
            listFiles = []
            cids = []

            # check
            if file_name_prefix == 'cid':
                # *** load only specific cids ***
                if len(file_name_ids) > 0 and len(list_file_names) > 0:
                    #
                    for i in range(len(list_file_names)):
                        _res = re.search(
                            r'(^cid_([0-9]+)(\.)([a-z0-9]*))', str(list_file_names[i]))
                        # check
                        if _res:
                            # save
                            _cid = _res.group(2)
                            cids.append(_cid)
                            # check
                            if _cid in file_name_ids:
                                _fileName = _res.group(1)
                                # save
                                listFiles.append(_fileName)
            elif file_name_prefix == 'file_name':
                # *** load only specific names ***
                if len(file_name_ids) > 0 and len(list_file_names) > 0:
                    #
                    _fileNames = [
                        item for item in file_name_ids if item in list_file_names]
                    # save
                    listFiles.extend(_fileName)
            # res
            return listFiles

        except Exception as e:
            raise Exception(e)

    @staticmethod
    def CreateCVS(csv_list, file_location, file_name='', file_format='json', save_location=''):
        '''
        Create csv file from dictionary

        Parameters
        ----------
        csv_list : list
            list of dictionary
        file_location : str
            file location
        file_name : str
            file name
        file_format : str
            file format
        save_location : str
            save location

        Returns
        -------
        _save_location : str
            save location
        csvFileName : str
            csv file name
        '''
        try:
            # check
            if len(save_location) == 0:
                _save_location = file_location
            else:
                _save_location = save_location

            # log
            _log = f"csv file is successfully created in {_save_location}"

            # csv file
            csvFileId = str(date.today()) + \
                "-" + str(random.randint(0, 256))

            # # file name
            if len(file_name) > 0:
                csvFileName = f'{file_name}.csv'
            else:
                csvFileName = f'cid_{csvFileId}.csv'

            # create path
            csvFile = os.path.join(_save_location, csvFileName)

            # write
            with open(csvFile, 'w') as f:
                writer = csv.DictWriter(
                    f, fieldnames=Utility.json_csv_columns)
                writer.writeheader()
                writer.writerows(csv_list)
                # log
                # print(_log)

            # res
            return _save_location, csvFileName
        except Exception as e:
            raise

    def ConvertCSVContentToList(csv_content):
        '''
        convert csv content (string) to list

        Parameters
        ----------
        csv_content : str
            csv string content

        Returns
        -------
        list
            list of csv content

        hint:
            list[0]: column head
            list[1]: records

        '''
        try:
            # csv column
            csv_column = []
            csv_rows = []

            #  csv list
            csv_list = csv_content.splitlines()
            # check
            if type(csv_list) is list:
                # len
                csv_list_len = len(csv_list)
                # csv column
                _column = str(csv_list[0]).replace(
                    "'", "").replace('"', '').split(',')
                csv_column.extend(_column)
                # csv rows
                _rows = str(csv_list[1]).replace(
                    "'", "").replace('"', '').split(',')
                csv_rows.extend(_rows)
            else:
                csv_list_len = -1

            # res
            return csv_list, csv_column, csv_rows, csv_list_len

        except Exception as e:
            raise Exception(e)

    @staticmethod
    def SaveNpArray(arr, file_name='', location=''):
        '''
        Save numpy array

        Parameters
        ----------
        arr : numpy array
            numpy array
        file_name : str
            file name
        location : str
            location

        '''
        try:
            # full file name
            if not file_name:
                file_name = "np_array"
            fullFileName = f"{file_name}.npy"
            # location
            fileLoc = os.path.join(location, fullFileName)

            # open
            with open(fileLoc, 'wb') as f:
                np.save(f, arr)

            print("save operation is done.")
        except Exception as e:
            raise Exception(e)

    @staticmethod
    def load_custom_functional_group(file_location) -> List[Dict[str, List[str]]]:
        '''
        load custom functional group

        Parameters
        ----------
        file_location : str
            file location

        Returns
        -------
        custom_functional_group : dict
            custom functional group

        '''
        try:
            if os.path.exists(file_location):
                # check yml file format
                with open(file_location, 'r') as f:
                    _ref = yaml.load(
                        f, Loader=yaml.FullLoader)
                # check not empty
                if _ref:
                    # ref keys
                    # _ref_keys = list(_ref.keys())

                    # custom functional group
                    custom_functional_groups: List[Dict[str, List[str]]] = []
                    for key, value in _ref.items():
                        # check value
                        if not value:
                            raise Exception(
                                "value is not found.")

                        # NOTE: check value format
                        if isinstance(value, list):
                            # save
                            custom_functional_groups.append(
                                {
                                    key: value
                                }
                            )
                        else:
                            raise Exception(
                                "value format is not valid.")

                    return custom_functional_groups
                else:
                    raise Exception("ref file is empty.")
            else:
                raise Exception("file path is not valid.")
        except Exception as e:
            raise Exception(e)



# GRAPHER
# ---------------
# import libs
import re
from typing import List, Dict, Union, Tuple, Optional


class Molecule():
    '''
    Build a molecule from a source
    '''

    def __init__(self, molecule_src, molecule_name):
        '''
        Initialize the molecule source

        Parameters
        ----------
        molecule_src : Dict[str,List[str]]
            The source of the molecule
        molecule_name : str
            The name of the molecule
        '''
        self.molecule_src = molecule_src
        self.molecule_name = molecule_name

    def build(self):
        '''
        Build a molecule from a source
        '''
        try:
            # SECTION: check molecule
            molecule_src_checked = self.check_molecule(self.molecule_src)

            # SECTION: construct molecule
            chain_info, molecule, constructed_molecule = self.construct_molecule(
                molecule_src_checked)

            # create molecule
            constructed_molecule_: Dict[str, List[str]] = {}
            constructed_molecule_[self.molecule_name] = constructed_molecule

            # SECTION: return
            return chain_info, molecule, constructed_molecule, constructed_molecule_
        except Exception as e:
            raise Exception(
                f"An error occurred while building the molecule: {e}")

    def extract_highest_index(self, chain):
        """
        Extracts the highest numerical index from atom labels in the given chain.

        Parameters
        ----------
        chain : list
            A list of strings representing bonds between atoms in a chain.

        Returns
        -------
        int
            The highest numerical index found in the chain.
        """
        indices = []
        for bond in chain:
            # Extract numbers from atoms
            atoms = re.findall(r'[A-Za-z]+(\d+)', bond)
            if '{' in bond or '}' in bond:
                continue  # Ignore bonds containing {Chain..}
            indices.extend(map(int, atoms))
        return max(indices) if indices else 0

    def search_for_main_chain(self, molecule_src: Dict[str, List[str]]) -> str:
        """
        Searches for the main chain in the molecule source.

        Parameters
        ----------
        molecule_src : dict
            A dictionary containing lists of strings representing bonds between atoms in a molecule.

        Returns
        -------
        str
            The name of the main chain found in the molecule source.

        Raises
        ------
        Exception
            If the main chain is not found in the molecule source.
        """
        try:
            for key, chain in molecule_src.items():
                # create a pattern to match the main chain
                pattern1 = re.compile(r'\{.*?\}\*.*')
                pattern2 = re.compile(r'.*\*\{.*?\}')

                # check if the main chain is found
                if any(pattern1.match(bond) or pattern2.match(bond) for bond in chain):
                    return str(key)

            # raise an exception if the main chain is not found
            raise Exception("Main chain not found in the molecule source.")
        except Exception as e:
            raise Exception(
                f"An error occurred while searching for the main chain: {e}")

    def check_molecule(self, molecule_src: Dict[str, List[str]]) -> Union[Dict[str, List[str]], bool]:
        """
        Checks if the molecule source is valid.

        Parameters
        ----------
        molecule_src : dict
            A dictionary containing lists of strings representing bonds between atoms in a molecule.

        Returns
        -------
        bool
            True if the molecule source is valid, False otherwise.
        """
        try:
            # create a copy of the molecule source
            molecule = {key: chain.copy()
                        for key, chain in molecule_src.items()}

            # checked molecule source
            molecule_src_checked = {}

            # check if the main chain is found
            main_chain = self.search_for_main_chain(molecule)
            if not main_chain:
                raise Exception("Main chain not found in the molecule source.")

            # add the main chain to the checked molecule source
            molecule_src_checked[main_chain] = molecule[main_chain]

            # main chain bonds
            main_chain_bonds = molecule[main_chain]

            # chain counter
            chain_counter = 0
            # looping through the molecule source
            for key, chain in molecule.items():
                if key != main_chain:
                    # check if the chain is connected to the main chain
                    # SECTION: create pattern to match the main chain
                    pattern_gate = rf"([A-Za-z]+)(\d+)\*\{{({re.escape(key)})\}}"

                    # looping through the main chain bonds
                    for i, bond in enumerate(main_chain_bonds):
                        match = re.match(pattern_gate, bond)
                        # check
                        if match:
                            # update chain counter
                            chain_counter += 1
                            # extract the matched chain
                            atom, index, key_chain = match.groups()
                            # rename key
                            key = key_chain + str(chain_counter)
                            # add the chain to the checked molecule source
                            molecule_src_checked[key] = chain

                            # update the element in the main chain
                            main_chain_bonds[i] = atom+index+"*{"+key+"}"

                    # reset chain counter
                    chain_counter = 0

            # res
            return molecule_src_checked
        except Exception as e:
            raise Exception(
                f"An error occurred while checking the molecule: {e}")

    def construct_molecule(self, molecule_src: Dict[str, List[str]]):
        """
        Constructs the molecule from the given molecule source.

        Parameters
        ----------
        molecule_src : dict
            A dictionary containing lists of strings representing bonds between atoms in a molecule.

        Returns
        -------
        dict
            A dictionary containing lists of strings representing bonds between atoms in a molecule.
        """
        try:
            # create a copy of the molecule source
            molecule = {
                key: chain.copy() for key, chain in molecule_src.items()}
            # print(f'molecule: {molecule}')

            # search for the main chain
            main_chain = self.search_for_main_chain(molecule)
            # print(f'main_chain: {main_chain}')

            chain_info: Dict[str, Dict[str, List[str]]] = {}

            # reset
            highest_index = 0

            # check if the main chain is found
            if main_chain:
                # get the highest index from the main chain
                highest_index = self.extract_highest_index(
                    molecule[main_chain])
                # print(f'highest_index: {highest_index}')

                # update index of other chains
                for key, chain in molecule.items():
                    if key != main_chain:
                        # create chain info
                        chain_info[key] = {
                            'bonds': [],
                            'gate': []
                        }

                        # update the index of the chain
                        for i, bond in enumerate(chain):

                            # SECTION: define pattern to match bonds
                            pattern = r"([A-Za-z]+)(\d+)([-=#])([A-Za-z]+)(\d+)"
                            # match the pattern
                            match_bond = re.match(pattern, bond)
                            # augment the index
                            if match_bond:
                                # extract atoms and indices
                                atom1, index1, bond_order, atom2, index2 = match_bond.groups()
                                # update the index
                                index1 = str(int(index1) + highest_index)
                                index2 = str(int(index2) + highest_index)
                                # update the bond
                                molecule[key][i] = f"{atom1}{index1}{bond_order}{atom2}{index2}"
                                # update the chain info
                                chain_info[key]['bonds'].append(
                                    molecule[key][i])

                            # SECTION: gate pattern
                            pattern_gate = r"([A-Za-z]+)(\d+)([-=#])\*"
                            # match the pattern
                            match_gate = re.match(pattern_gate, bond)
                            # augment the index
                            if match_gate:
                                # extract atoms and indices
                                atom1, index1, bond_order = match_gate.groups()
                                # update the index
                                index1 = str(int(index1) + highest_index)
                                # update the bond
                                molecule[key][i] = f"{bond_order}{atom1}{index1}"
                                # update the chain info
                                chain_info[key]['gate'].append(
                                    molecule[key][i])

                            # SECTION: gate pattern
                            pattern_gate = r"\*([-=#])([A-Za-z]+)(\d+)"
                            # match the pattern
                            match_gate = re.match(pattern_gate, bond)
                            # augment the index
                            if match_gate:
                                # extract atoms and indices
                                bond_order, atom1, index1 = match_gate.groups()
                                # update the index
                                index1 = str(int(index1) + highest_index)
                                # update the bond
                                molecule[key][i] = f"{atom1}{index1}{bond_order}"
                                # update the chain info
                                chain_info[key]['gate'].append(
                                    molecule[key][i])

                        # update highest index
                        highest_index = int(index1)
                        # print(f"highest_index: {highest_index}")

            # combine the main chain and other chains
            constructed_molecule: List[str] = []
            # find the gate atoms
            for items in molecule[main_chain]:
                # SECTION: define pattern
                pattern_gate = r"([A-Za-z]+)(\d+)\*\{([A-Za-z0-9]+)\}"
                # match the pattern
                match_gate = re.match(pattern_gate, items)
                # extract the gate atoms
                if match_gate:
                    # extract atoms and indices
                    atom1, index1, gate = match_gate.groups()
                    # find element index in molecule['main_chain']
                    element_index = molecule[main_chain].index(items)
                    # update
                    # check start with letter or number
                    if chain_info[gate]['gate'][0].startswith(('-', '=', '#')):
                        molecule[main_chain][element_index] = f"{atom1}{index1}{chain_info[gate]['gate'][0]}"
                    else:
                        molecule[main_chain][element_index] = f"{chain_info[gate]['gate'][0]}{atom1}{index1}"

                pattern_gate = r"\{([A-Za-z0-9]+)\}\*([A-Za-z]+)(\d+)"
                # match the pattern
                match_gate = re.match(pattern_gate, items)
                # extract the gate atoms
                if match_gate:
                    # extract atoms and indices
                    gate, atom1, index1 = match_gate.groups()
                    # find element index in molecule['main_chain']
                    element_index = molecule[main_chain].index(items)
                    # update
                    # check start with letter or number
                    if chain_info[gate]['gate'][0].startswith(('-', '=', '#')):
                        molecule[main_chain][element_index] = f"{atom1}{index1}{chain_info[gate]['gate'][0]}"
                    else:
                        molecule[main_chain][element_index] = f"{chain_info[gate]['gate'][0]}{atom1}{index1}"

            # combine the main chain and other chains
            for key, chain in molecule.items():
                if key == main_chain:
                    constructed_molecule.extend(chain)

            # chain info
            for key, chain in chain_info.items():
                constructed_molecule.extend(chain['bonds'])

            return chain_info, molecule, constructed_molecule
        except Exception as e:
            raise Exception(
                f"An error occurred while constructing the molecule: {e}")
